#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <cuda.h>
#include "defs.h"

/*
 *  C Interface functions from kernel.cu
 */
extern float applySignalCPU(Matrix, Signals, Signals);
extern float applySignalGPU(Matrix, Signals, Signals);

extern int checkResults(Signals, Signals);

// random fill for the transformation matrix
Matrix randfillmatrix()
{
    int i;
    Matrix transMat;

    transMat = (Matrix) malloc(sizeof(Matrix) * MAT_ROW_DIM * MAT_COL_DIM);

    if(transMat == NULL){
        ERROR_MSG("error : matrix memory allocation failed\n")
    }


    for(i = 0; i < MAT_ROW_DIM * MAT_COL_DIM; i++) {
	transMat[i] = (float)(rand()/(float) RAND_MAX);
    }
    
    return transMat;
}

// allocates array space to contain all the signals
 
Signals allocatesignalseq(int nseq)
{
    Signals seqVec;
    seqVec = (Signals) malloc(sizeof(float) * nseq * VEC_DIM);

    if(seqVec == NULL) {
        ERROR_MSG("error: memory allocation failed for signal array\n")
    }

    return seqVec;
}

// allocates memory for each signal and fill rand value
void randfillsignalseq(Signals signals, int nseq)
{
    int i;
    int size = VEC_DIM * nseq;
    for(i = 0; i < size; i++) {
	signals[i] = (float)(rand()/(float) RAND_MAX);
    }
}


// Note: Use this function only for smaller matrix dimension
void dbgPrintMatrix(Matrix Mat)
{
    int i,j;
    printf("Matrix:\n");
    for(i = 0; i < MAT_ROW_DIM; i++)
    {        
        for(j = 0; j < MAT_COL_DIM; j++)
        {
            printf("%.3f ", Mat[i*MAT_COL_DIM + j]);
        }
        printf("\n");
    }

    printf("\n");
}

// Note: Use this function only for smaller signal dimension
void dbgPrintSignals(Signals signals, int nseq, int signalsize)
{
    int i;
    printf("Signal Vectors:\n");
    printf("0: ");
    // print all the signals
    for(i = 0; i < nseq * signalsize; )
    {
        printf("%.3f ", signals[i]);
        i++;
        if(i >= VEC_DIM && !(i%VEC_DIM))
            printf("\n%d: ", i/VEC_DIM);
    }

    printf("\n");
}


int main()
{
    int i;
    srand(time(NULL));
    
    // allocate and fill matrix
    Matrix _transMatrix = randfillmatrix();

    // allocate one large array for all signal vectors
    Signals  signals =  allocatesignalseq(NSEQ);

    // allocate and fill rand values for each signal
    randfillsignalseq(signals, NSEQ);

    // allocate one large array for all the resultant signal vectors
    // result dimension : MAT_ROW_DIM x 1
    Signals _cpuresult = (float*) malloc(sizeof(float) * MAT_ROW_DIM * NSEQ);
    
   // copy the results from GPU back to this variable
   // result dimension : MAT_ROW_DIM x 1
    float *_gpuresult = (float*) malloc(sizeof(float) * RES_MAT_ROW * RES_MAT_COL);

    float seq_elapsed_time, cuda_elapsed_time;

    seq_elapsed_time = applySignalCPU(_transMatrix, signals, _cpuresult);

    printf("Returned from CPU kernel");
    int row, col;
    //Transpose of _ipsignal
     Signals B_T = allocatesignalseq(10);
     for(row = 0; row < 5120; row++){
	  for(col = 0; col < 10; col++){
		B_T[col + row * NSEQ] = signals[VEC_DIM * col + row];	
		}
    }

    cuda_elapsed_time = applySignalGPU(_transMatrix, B_T, _gpuresult);

    //Transpose of _cpuresult
     Signals _cpuResult = allocatesignalseq(10);
     for(row = 0; row < 5120; row++){
	  for(col = 0; col < 10; col++){
		_cpuResult[col + row * NSEQ] = _cpuresult[VEC_DIM * col + row];	
		}
    }

    printf("Test %s\n",checkResults(_cpuResult, _gpuresult) ? "Passed":"Failed");
    printf("Seq. Elapsed Time:\t%.2f ms\n", seq_elapsed_time);
    printf("CUDA Elapsed Time:\t%.2f ms\n", cuda_elapsed_time);

     // free memory allocated
   // free(_transMatrix);
   // free(signals);
   // free(B_T);
   // free(_gpuresult);
   // free(_cpuresult);
   
    return 0;
}

